import royalnet.utils as ru
import royalnet.backpack.tables as rbt
import royalnet.constellation.api as rca
import royalnet.constellation.api.apierrors as rcae
import itsdangerous
import aiohttp
import aiohttp.client_exceptions
import datetime
from ..types import oauth_refresh
from ..tables import Osu, FiorygiTransaction


class ApiAuthLoginOsuStar(rca.ApiStar):
    path = "/api/auth/login/osu/v1"

    parameters = {
        "get": {
            "code": "The code returned by the osu! API.",
            "state": "(Optional) The state payload generated by the osu! command to link a new account. "
                     "If missing, just login."
        }
    }

    auth = {
        "get": False,
    }

    tags = ["auth"]

    @property
    def client_id(self):
        return self.config['osu']['client_id']

    @property
    def client_secret(self):
        return self.config['osu']['client_secret']

    @property
    def base_url(self):
        return self.config['base_url']

    @property
    def secret_key(self):
        return self.config['secret_key']

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Login to Royalnet with your osu! account."""
        OsuT = self.alchemy.get(Osu)
        TokenT = self.alchemy.get(rbt.Token)

        code = data.str("code")
        state = data.str("state", optional=True)

        if state is not None:
            serializer = itsdangerous.URLSafeSerializer(self.config["secret_key"], salt="osu")
            uid = serializer.loads(state)
            user = await rbt.User.find(self.alchemy, data.session, uid)
        else:
            user = None

        try:
            t = await oauth_refresh(url="https://osu.ppy.sh/oauth/token",
                                    client_id=self.client_id,
                                    client_secret=self.client_secret,
                                    redirect_uri=f"{self.base_url}{self.path}",
                                    refresh_code=code)
        except aiohttp.client_exceptions.ClientResponseError:
            raise rca.ForbiddenError("osu! API returned an error in the OAuth token exchange")

        async with aiohttp.ClientSession(headers={"Authorization": f"Bearer {t['access_token']}"}) as session:
            async with session.get("https://osu.ppy.sh/api/v2/me/") as response:
                m = await response.json()

        if user is not None:
            osu = OsuT(
                user=user,
                access_token=t["access_token"],
                refresh_token=t["refresh_token"],
                expiration_date=datetime.datetime.now() + datetime.timedelta(seconds=t["expires_in"]),
                osu_id=m["id"],
                username=m["username"]
            )

            data.session.add(osu)
        else:
            osu = await ru.asyncify(
                data.session.query(OsuT).filter_by(osu_id=m["id"]).all
            )
            if osu is None:
                raise rcae.ForbiddenError("Unknown osu! account")
            user = osu.user

        if self.config["osu"]["login"]["enabled"]:
            token: rbt.Token = TokenT.generate(alchemy=self.alchemy, user=user, expiration_delta=datetime.timedelta(days=7))
            data.session.add(token)
            await data.session_commit()

            return token.json()
        else:
            raise rcae.ForbiddenError("Account linked successfully; cannot use this account to generate a Royalnet"
                                      " login token, as osu! login is currently disabled on this Royalnet instance.")


from typing import *
import royalnet
import royalnet.commands as rc
import random
import datetime


class AnswerCommand(rc.Command):
    name: str = "answer"

    description: str = "Fai una domanda al bot, che possa essere risposta con un sì o un no: lui ti risponderà!"

    syntax: str = ""

    _answers = [
        #Cerchiamo di tenere bilanciate le tre colonne, o almeno le prime due.
        #Se avete un'idea ma metterebbe troppe opzioni in un'unica categoria, mettetela sotto commento.

        #risposte "sì"
        "Sì.",
        "Decisamente sì!",
        "Uhm, secondo me sì.",
        "Sì! Sì! SÌ!",
        "Yup.",
        "👍",
        "Direi proprio di sì.",
        "Assolutamente sì.",
        "Ma certo!",
        "✔️",
        "👌",
        "Esatto!",
        "Senz'altro!",
        "Ovviamente.",
        "Questa domanda ha risposta affermativa.",
        "Hell yeah.",

        #risposte "no"
        "No.",
        "Decisamente no!",
        "Uhm, secondo me sì.",
        "No, no, e ancora NO!",
        "Nope.",
        "👎",
        "Direi proprio di no.",
        "Assolutamente no.",
        "Certo che no!",
        "✖️",
        "🙅",
        "Neanche per idea!",
        "Neanche per sogno!",
        "Niente affatto!",
        "Questa domanda ha risposta negativa.",
        "Hell no.",

        #risposte "boh"
        "Boh.",
        "E io che ne so?!",
        "Non so proprio rispondere",
        "Non lo so",
        "Mi rifiuto di rispondere alla domanda!",
        "Non parlerò senza il mio avvocato!",
        "Dunno.",
        "Perché lo chiedi a me?",
        "🤷 Ah, non lo so io! ¯\_(ツ)_/¯",
        "🤷",
        "¯\_(ツ)_/¯",
        "No idea.",
        "Dunno.",
        "Boooooh!",
        "Non ne ho la più pallida idea.",
    ]

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:

        h = hash(datetime.datetime.now())

        r = random.Random(x=h)

        message = r.sample(self._answers, 1)[0]
        await data.reply(message)





from typing import *
import royalnet.commands as rc
import royalnet.utils as ru
import royalnet.serf.telegram as rst
import royalnet.backpack.tables as rbt
import abc
import logging
import asyncio as aio
from ...types import Updatable


log = logging.getLogger(__name__)


class LinkerCommand(rc.Command, metaclass=abc.ABCMeta):

    def __init__(self, interface: rc.CommandInterface):
        super().__init__(interface)

        self.updater_task = None
        if self.enabled():
            self.updater_task = self.loop.create_task(self.run_updater())

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        author = await data.get_author(error_if_none=True)
        if len(args) == 0:
            message = []
            for obj in await self.get_updatables_of_user(session=data.session, user=author):
                async def change(attribute: str, value: Any):
                    """A shortcut for self.__change."""
                    await self._change(session=data.session,
                                       obj=obj,
                                       attribute=attribute,
                                       new=value)

                await self.update(session=data.session, obj=obj, change=change)
                message.append(self.describe(obj))
            if len(message) == 0:
                raise rc.UserError("Nessun account connesso.")
            await data.session_commit()
            await data.reply("\n".join(message))
        else:
            created = await self.create(session=data.session, user=author, args=args, data=data)
            await data.session_commit()
            if created is not None:
                message = ["🔗 Account collegato!", "", self.describe(created)]
                await data.reply("\n".join(message))

    def describe(self, obj: Updatable) -> str:
        """The text that should be appended to the report message for a given Updatable."""
        return str(obj)

    @abc.abstractmethod
    async def get_updatables_of_user(self, session, user: rbt.User) -> List[Updatable]:
        """Get the updatables of a specific user."""
        ...

    @abc.abstractmethod
    async def get_updatables(self, session) -> List[Updatable]:
        """Return a list of all objects that should be updated at this updater cycle."""
        ...

    @abc.abstractmethod
    async def create(self,
                     session,
                     user: rbt.User,
                     args: rc.CommandArgs,
                     data: Optional[rc.CommandData] = None) -> Optional[Updatable]:
        """Create a new updatable object for a user.

        This function is responsible for adding the object to the session."""
        ...

    @abc.abstractmethod
    async def update(self, session, obj, change: Callable[[str, Any], Awaitable[None]]):
        """Update a single updatable object. Use the change method to change values on the object!"""
        ...

    @abc.abstractmethod
    async def on_increase(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        """Called when the attribute has increased from the old value."""
        ...

    @abc.abstractmethod
    async def on_unchanged(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        """Called when the attribute stayed the same as the old value."""
        ...

    @abc.abstractmethod
    async def on_decrease(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        """Called when the attribute has decreased from the old value."""
        ...

    @abc.abstractmethod
    async def on_first(self, session, obj: Updatable, attribute: str, old: None, new: Any) -> None:
        """Called when the attribute changed from None."""
        ...

    @abc.abstractmethod
    async def on_reset(self, session, obj: Updatable, attribute: str, old: Any, new: None) -> None:
        """Called when the attribute changed to None."""
        ...

    async def _change(self,
                      session,
                      obj,
                      attribute: str,
                      new) -> None:
        """Set the value of an attribute of an object to a value, and call the corresponding method."""
        old = obj.__getattribute__(attribute)
        if new == old:
            await self.on_unchanged(session=session,
                                    obj=obj,
                                    attribute=attribute,
                                    old=old,
                                    new=new)
        else:
            if old is None:
                await self.on_first(session=session,
                                    obj=obj,
                                    attribute=attribute,
                                    old=old,
                                    new=new)
            elif new is None:
                await self.on_reset(session=session,
                                    obj=obj,
                                    attribute=attribute,
                                    old=old,
                                    new=new)
            elif new > old:
                await self.on_increase(session=session,
                                       obj=obj,
                                       attribute=attribute,
                                       old=old,
                                       new=new)
            else:
                await self.on_decrease(session=session,
                                       obj=obj,
                                       attribute=attribute,
                                       old=old,
                                       new=new)
        obj.__setattr__(attribute, new)

    def enabled(self) -> bool:
        """Whether the updater is enabled or not."""
        return self.config[self.name]["updater"]["enabled"] and self.interface.name == "telegram"

    def period(self) -> int:
        """The time between two updater cycles."""
        return self.config[self.name]["updater"]["period"]

    def delay(self) -> int:
        """The time between two object updates."""
        return self.config[self.name]["updater"]["delay"]

    def target(self) -> int:
        """The id of the Telegram chat where notifications should be sent."""
        return self.config[self.name]["updater"]["target"]

    async def run_updater(self):
        log.info(f"Starting updater: {self.name}")

        while True:
            log.debug(f"Updater cycle: {self.name}")
            session = self.alchemy.Session()
            objects = await self.get_updatables(session)

            for obj in objects:
                log.debug(f"Updating: {obj} ({self.name})")

                async def change(attribute: str, value: Any):
                    """A shortcut for self.__change."""
                    await self._change(session=session,
                                       obj=obj,
                                       attribute=attribute,
                                       new=value)

                try:
                    await self.update(session=session,
                                      obj=obj,
                                      change=change)
                except Exception as e:
                    ru.sentry_exc(e)

                delay = self.delay()
                log.debug(f"Waiting for: {delay} seconds (delay)")
                await aio.sleep(delay)

            log.debug(f"Committing updates: {self.name}")
            await ru.asyncify(session.commit)
            session.close()

            period = self.period()
            log.debug(f"Waiting for: {period} seconds (period)")
            await aio.sleep(period)

    async def notify(self, message):
        await self.serf.api_call(self.serf.client.send_message,
                                 chat_id=self.target(),
                                 text=rst.escape(message),
                                 parse_mode="HTML",
                                 disable_webpage_preview=True)



from typing import *
import royalnet
import royalnet.commands as rc


class HelpCommand(rc.Command):
    name: str = "help"

    description: str = "Visualizza informazioni su un comando."

    syntax: str = "{comando}"

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        if len(args) == 0:
            message = [
                "ℹ️ Comandi disponibili:"
            ]

            for command in sorted(list(set(self.serf.commands.values())), key=lambda c: c.name):
                message.append(f"- [c]{self.interface.prefix}{command.name}[/c]")

            await data.reply("\n".join(message))
        else:
            name: str = args[0].lstrip(self.interface.prefix)

            try:
                command: rc.Command = self.serf.commands[f"{self.interface.prefix}{name}"]
            except KeyError:
                raise rc.InvalidInputError("Il comando richiesto non esiste.")

            message = [
                f"ℹ️ [c]{self.interface.prefix}{command.name} {command.syntax}[/c]",
                "",
                f"{command.description}"
            ]

            await data.reply("\n".join(message))



from typing import *
import itsdangerous
import aiohttp

from royalnet.backpack import tables as rbt
import royalnet.commands as rc
import royalnet.utils as ru

from .abstract.linker import LinkerCommand
from ..types import Updatable
from ..tables import Osu
from ..stars.api_auth_login_osu import ApiAuthLoginOsuStar


class OsuCommand(LinkerCommand):
    name = "osu"

    description = "Connetti e sincronizza il tuo account di osu!"

    @property
    def client_id(self):
        return self.config[self.name]['client_id']

    @property
    def client_secret(self):
        return self.config[self.name]['client_secret']

    @property
    def base_url(self):
        return self.config['base_url']

    @property
    def secret_key(self):
        return self.config['secret_key']

    async def get_updatables_of_user(self, session, user: rbt.User) -> List[Osu]:
        return user.osu

    async def get_updatables(self, session) -> List[Osu]:
        return await ru.asyncify(session.query(self.alchemy.get(Osu)).all)

    async def create(self,
                     session,
                     user: rbt.User,
                     args: rc.CommandArgs,
                     data: Optional[rc.CommandData] = None) -> Optional[Osu]:
        serializer = itsdangerous.URLSafeSerializer(self.secret_key, salt="osu")
        # TODO: Ensure the chat the link is being sent in is secure!!!
        await data.reply("🔑 [b]Login necessario[/b]\n"
                         f"[url=https://osu.ppy.sh/oauth/authorize"
                         f"?client_id={self.client_id}"
                         f"&redirect_uri={self.base_url}{ApiAuthLoginOsuStar.path}"
                         f"&response_type=code"
                         f"&state={serializer.dumps(user.uid)}]"
                         f"Connetti account di osu! a {user.username}"
                         f"[/url]")
        return None

    async def update(self, session, obj: Osu, change: Callable[[str, Any], Awaitable[None]]):
        await obj.refresh_if_expired(client_id=self.client_id,
                                     client_secret=self.client_secret,
                                     base_url=self.base_url,
                                     path=ApiAuthLoginOsuStar.path)
        async with aiohttp.ClientSession(headers={"Authorization": f"Bearer {obj.access_token}"}) as session:
            async with session.get("https://osu.ppy.sh/api/v2/me/osu") as response:
                m = await response.json()
                obj.avatar_url = m["avatar_url"]
                obj.username = m["username"]
                if "statistics" in m:
                    await change("standard_pp", m["statistics"].get("pp"))
            async with session.get("https://osu.ppy.sh/api/v2/me/taiko") as response:
                m = await response.json()
                if "statistics" in m:
                    await change("taiko_pp", m["statistics"].get("pp"))
            async with session.get("https://osu.ppy.sh/api/v2/me/fruits") as response:
                m = await response.json()
                if "statistics" in m:
                    await change("catch_pp", m["statistics"].get("pp"))
            async with session.get("https://osu.ppy.sh/api/v2/me/mania") as response:
                m = await response.json()
                if "statistics" in m:
                    await change("mania_pp", m["statistics"].get("pp"))

    async def on_increase(self, session, obj: Osu, attribute: str, old: Any, new: Any) -> None:
        if attribute == "standard_pp":
            await self.notify(f"📈 [b]{obj.user}[/b] è salito a [b]{new:.0f}pp[/b] su [i]osu![/i]! Congratulazioni!")
        elif attribute == "taiko_pp":
            await self.notify(f"📈 [b]{obj.user}[/b] è salito a [b]{new:.0f}pp[/b] su [i]osu!taiko[/i]! Congratulazioni!")
        elif attribute == "catch_pp":
            await self.notify(f"📈 [b]{obj.user}[/b] è salito a [b]{new:.0f}pp[/b] su [i]osu!catch[/i]! Congratulazioni!")
        elif attribute == "mania_pp":
            await self.notify(f"📈 [b]{obj.user}[/b] è salito a [b]{new:.0f}pp[/b] su [i]osu!mania[/i]! Congratulazioni!")

    async def on_unchanged(self, session, obj: Osu, attribute: str, old: Any, new: Any) -> None:
        pass

    async def on_decrease(self, session, obj: Osu, attribute: str, old: Any, new: Any) -> None:
        if attribute == "standard_pp":
            await self.notify(f"📉 [b]{obj.user}[/b] è sceso a [b]{new:.0f}pp[/b] su [i]osu![/i].")
        elif attribute == "taiko_pp":
            await self.notify(f"📉 [b]{obj.user}[/b] è sceso a [b]{new:.0f}pp[/b] su [i]osu!taiko[/i].")
        elif attribute == "catch_pp":
            await self.notify(f"📉 [b]{obj.user}[/b] è sceso a [b]{new:.0f}pp[/b] su [i]osu!catch[/i].")
        elif attribute == "mania_pp":
            await self.notify(f"📉 [b]{obj.user}[/b] è sceso a [b]{new:.0f}pp[/b] su [i]osu!mania[/i].")

    async def on_first(self, session, obj: Osu, attribute: str, old: None, new: Any) -> None:
        if attribute == "standard_pp":
            await self.notify(f"⭐️ [b]{obj.user}[/b] ha guadagnato i suoi primi [b]{new:.0f}pp[/b] su [i]osu![/i]!")
        elif attribute == "taiko_pp":
            await self.notify(f"⭐️ [b]{obj.user}[/b] ha guadagnato i suoi primi [b]{new:.0f}pp[/b] su [i]osu!taiko[/i]!")
        elif attribute == "catch_pp":
            await self.notify(f"⭐️ [b]{obj.user}[/b] ha guadagnato i suoi primi [b]{new:.0f}pp[/b] su [i]osu!catch[/i]!")
        elif attribute == "mania_pp":
            await self.notify(f"⭐️ [b]{obj.user}[/b] ha guadagnato i suoi primi [b]{new:.0f}pp[/b] su [i]osu!mania[/i]!")

    async def on_reset(self, session, obj: Osu, attribute: str, old: Any, new: None) -> None:
        if attribute == "standard_pp":
            await self.notify(f"⬜️ [b]{obj.user}[/b] non è più classificato su [i]osu![/i].")
        elif attribute == "taiko_pp":
            await self.notify(f" ⬜️[b]{obj.user}[/b] non è più classificato su [i]osu!taiko[/i].")
        elif attribute == "catch_pp":
            await self.notify(f"⬜️ [b]{obj.user}[/b] non è più classificato su [i]osu!catch[/i].")
        elif attribute == "mania_pp":
            await self.notify(f"⬜️ [b]{obj.user}[/b] non è più classificato su [i]osu!mania[/i].")



from typing import *
import royalnet.commands as rc
import aiohttp
import io


class CatCommand(rc.Command):
    name: str = "cat"

    description: str = "Invia un gatto casuale in chat."

    syntax: str = ""

    aliases = ["catto", "kat", "kitty", "kitten", "gatto", "miao", "garf", "basta"]

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get("https://api.thecatapi.com/v1/images/search") as response:
                if response.status >= 400:
                    raise rc.ExternalError(f"Request returned {response.status}")
                result = await response.json()
                assert len(result) == 1
                cat = result[0]
                assert "url" in cat
                url = cat["url"]
            async with session.get(url) as response:
                img = await response.content.read()
                await data.reply_image(image=io.BytesIO(img))



from typing import *
import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables import Poll


class ApiPollsListStar(rca.ApiStar):
    path = "/api/poll/list/v2"

    tags = ["poll"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get a list of all polls."""
        PollT = self.alchemy.get(Poll)

        polls: List[Poll] = await ru.asyncify(data.session.query(PollT).all)

        return list(map(lambda p: {
            "id": p.id,
            "question": p.question,
            "creator": p.creator.json(),
            "expires": p.expires.isoformat(),
            "created": p.created.isoformat(),
        }, polls))



from typing import *
import royalnet.commands as rc
import royalnet.backpack.tables as rbt


class ExecCommand(rc.Command):
    # oh god if there is a security vulnerability
    name: str = "exec"

    description: str = "Esegui uno script Python... se sei Steffo."

    syntax: str = "{script}"

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        user: rbt.User = await data.get_author(error_if_none=True)
        if "admin" not in user.roles:
            raise rc.CommandError("Non sei autorizzato a eseguire codice arbitrario!\n"
                               "(Sarebbe un po' pericoloso se te lo lasciassi eseguire, non trovi?)")
        try:
            exec(args.joined(require_at_least=1))
        except Exception as e:
            raise rc.CommandError(f"Esecuzione fallita: {e}")
        await data.reply(f"✅ Fatto!")



import discord
from typing import *
import royalnet.commands as rc
from royalnet.serf.discord import *


class DiscordPauseEvent(rc.Event):
    name = "discord_pause"

    async def run(self,
                  guild_id: Optional[int] = None,
                  **kwargs) -> dict:
        if not isinstance(self.serf, DiscordSerf):
            raise rc.UnsupportedError()
        client: discord.Client = self.serf.client
        if len(self.serf.voice_players) == 1:
            voice_player: VoicePlayer = self.serf.voice_players[0]
        else:
            if guild_id is None:
                # TODO: trovare un modo per riprodurre canzoni su più server da Telegram
                raise rc.InvalidInputError("Non so in che Server riprodurre questo file...\n"
                                           "Invia il comando su Discord, per favore!")
            guild: discord.Guild = client.get_guild(guild_id)
            if guild is None:
                raise rc.InvalidInputError("Impossibile trovare il Server specificato.")
            candidate_players = self.serf.find_voice_players(guild)
            if len(candidate_players) == 0:
                raise rc.UserError("Il bot non è in nessun canale vocale.\n"
                                   "Evocalo prima con [c]summon[/c]!")
            elif len(candidate_players) == 1:
                voice_player = candidate_players[0]
            else:
                raise rc.CommandError("Non so su che Server saltare canzone...\n"
                                      "Invia il comando su Discord, per favore!")

        if voice_player.voice_client.is_paused():
            voice_player.voice_client.resume()
            return {"action": "resumed"}
        else:
            voice_player.voice_client.pause()
            return {"action": "paused"}



from .mmchoice import MMChoice
from .mminterfacedata import MMInterfaceData, MMInterfaceDataTelegram
from .leaguetier import LeagueTier
from .leaguerank import LeagueRank
from .leagueleague import LeagueLeague
from .dotamedal import DotaMedal
from .dotastars import DotaStars
from .dotarank import DotaRank
from .brawlhallatier import BrawlhallaTier
from .brawlhallametal import BrawlhallaMetal
from .brawlhallarank import BrawlhallaRank
from .pollmood import PollMood
from .updatable import Updatable
from .oauth_refresh import oauth_refresh


__all__ = [
    "MMChoice",
    "MMInterfaceData",
    "MMInterfaceDataTelegram",
    "LeagueTier",
    "LeagueRank",
    "LeagueLeague",
    "DotaMedal",
    "DotaStars",
    "DotaRank",
    "BrawlhallaMetal",
    "BrawlhallaRank",
    "BrawlhallaTier",
    "PollMood",
    "Updatable",
    "oauth_refresh",
]
